import { Sort } from '../sort.mjs';
// import { CountingSort } from '../counting-sort/counting-sort.mjs';

export class RadixSort extends Sort {
  /**
   * lsd
   * stable
   * in-place
   * Sorts an array using "radix sort"
   *
   * @param {Number[]} array - array to sort
   * @param {Number} [radix=10] - the base/radix to use
   * @return {Number[]} array - sorted array
   *
   */
  sort(array, radix = 10) {
    if (!array.length) {
      return array;
    }

    //** Determine minimum and maximum values **//
    const min = Math.min(...array);
    const max = Math.max(...array);

    /* Perform counting sort on each exponent/digit */
    /* Starting at the least significant digit */
    let exponent = 1;

    while ((max - min) / exponent >= 1) {
      array = this.countingSortByDigit(array, radix, exponent, min);
      exponent *= radix;
    }

    return array;
  }

  /**
   * stable
   * sorts an array by a particular digit using "counting sort"
   *
   * @param {Number[]} array - array to sort
   * @param {Number} raxix - the base/radix to use to sort
   * @param {Number} exponent - the exponent of the significant digit to sort
   * @param {Number} minValue - the minimum value within the array
   * @return {Number[]} array - sorted array
   */
  countingSortByDigit(array, radix, exponent, minValue) {
    const output = new Array(array.length);
    const buckets = new Array(radix).fill(0);
    let bucketIndex;

    /* Count frequencies */
    for (let i = 0; i < array.length; i++) {
      bucketIndex = Math.floor(((array[i] - minValue) / exponent) % radix);
      buckets[bucketIndex]++;
    }

    /* Compute cumulates */
    for (let i = 1; i < radix; i++) {
      buckets[i] += buckets[i - 1];
    }

    /* Move records */
    for (let i = array.length - 1; i >= 0; i--) {
      bucketIndex = Math.floor(((array[i] - minValue) / exponent) % radix);
      output[--buckets[bucketIndex]] = array[i];
    }

    /* Copy back */
    for (let i = 0; i < array.length; i++) {
      array[i] = output[i];
    }

    return array;
  }
}
