import { Comparator } from '../../comparator/comparator.mjs';

/**
 * Parent class for Min and Max Heaps
 */
export class Heap {
  /**
   * @constructs Heap
   * @param {Function} [comparatorFunction]
   */
  constructor(comparatorFunction) {
    if (new.target === Heap) {
      // throw new TypeError('Cannot construct Heap instance directly');
    }

    // Array representation of the heap
    this._heap = [];
    this.compare = new Comparator(comparatorFunction);
  }

  /**
   * @param {Number} childIndex
   * @return {Number}
   */
  getParentIndex(childIndex) {
    return Math.floor((childIndex - 1) / 2);
  }

  /**
   * @param {Number} parentIndex
   * @return {Number}
   */
  getLeftChildIndex(parentIndex) {
    return parentIndex * 2 + 1;
  }

  /**
   * @param {Number} parentIndex
   * @return {Number}
   */
  getRightChildIndex(parentIndex) {
    return parentIndex * 2 + 2;
  }

  /**
   * @param {Number} childIndex
   * @return {Boolean}
   */
  hasParent(childIndex) {
    return this.getParentIndex(childIndex) >= 0;
  }

  /**
   * @param {Number} parentIndex
   * @return {Boolean}
   */
  hasLeftChild(parentIndex) {
    return this.getLeftChildIndex(parentIndex) < this._heap.length;
  }

  /**
   * @param {Number} parentIndex
   * @return {Boolean}
   */
  hasRightChild(parentIndex) {
    return this.getRightChildIndex(parentIndex) < this._heap.length;
  }


  /**
   * @param {Number} parentIndex
   * @return {*}
   */
  leftChild(parentIndex) {
    return this._heap[this.getLeftChildIndex(parentIndex)];
  }

  /**
   * @param {Number} parentIndex
   * @return {*}
   */
  rightChild(parentIndex) {
    return this._heap[this.getRightChildIndex(parentIndex)];
  }

  /**
   * @param {Number} childIndex
   * @return {*}
   */
  parent(childIndex) {
    return this._heap[this.getParentIndex(childIndex)];
  }

  /**
   * @param {Number} indexOne
   * @param {Number} indexTwo
   */
  swap(indexOne, indexTwo) {
    [this._heap[indexOne], this._heap[indexTwo]] = [this._heap[indexTwo], this._heap[indexOne]];
  }

  /**
   * @return {*}
   */
  peek() {
    if (this._heap.length === 0) {
      return null;
    }
    return this._heap[0];
  }

  /**
   * @return {*}
   */
  poll() {
    if (this._heap.length === 0) {
      return null;
    }

    if (this._heap.length === 1) {
      return this._heap.pop();
    }

    const item = this._heap[0];

    /* Move the last element from the end to the head. */
    this._heap[0] = this._heap.pop();
    this.heapifyDown();

    return item;
  }

  /**
   * @param {*} item
   * @return {Heap}
   */
  add(item) {
    this._heap.push(item);
    this.heapifyUp();
    return this;
  }

  /**
   * @param {*} item
   * @param {Comparator} comparator
   * @return {Heap}
   */
  remove(item, comparator = this.compare) {
    /* Find count of items to remove */
    const count = this.find(item, comparator).length;

    for (let i = 0; i < count; i++) {
      /**
       * We need to find item index to remove each time after removal
       * since indices are being changed after each heapify (up/down) process
       */
      const indexToRemove = this.find(item).pop();

      /* If we need to remove last child in the heap then just remove it */
      /* There is no need to heapify the heap afterwards */
      if (indexToRemove === this._heap.length - 1) {
        this._heap.pop();
      }
      else {
        /* Move last element in heap to vacant (removed) position */
        this._heap[indexToRemove] = this._heap.pop();

        /* Get parent */
        const parentItem = this.parent(indexToRemove);

        /**
         * If there is no parent or parent is in correct order with the node
         * we are going to delete then heapify down. Otherwise heapify up.
         */
        if (this.hasLeftChild(indexToRemove) && (!parentItem || this.pairIsInCorrectOrder(parentItem, this._heap[indexToRemove]))) {
          this.heapifyDown(indexToRemove);
        }
        else {
          this.heapifyUp(indexToRemove);
        }
      }
    }

    return this;
  }

  /**
   * @param {*} item
   * @param {Comparator} comparator
   * @return {Number[]}
   */
  find(item, comparator = this.compare) {
    return this._heap.reduce((res, it, idx) => {
      if (comparator.eq(item, it)) res.push(idx);
      return res;
    }, []);
  }

  /**
   * @return {Boolean}
   */
  isEmpty() {
    return !this._heap.length;
  }

  /**
   * @return {String}
   */
  toString() {
    return this._heap.toString();
  }

  /**
   * @return {Array}
   */
  toArray() {
    return this._heap;
  }

  /**
   * @param {Number} customStartIndex
   */
  heapifyUp(customStartIndex) {
    /**
     * Take last element and lift it up
     * until it is in the correct order
     * with respect to its parent element
     */
    let currentIndex = customStartIndex || this._heap.length - 1;
    const condition = (index) => this.hasParent(index) && !this.pairIsInCorrectOrder(this.parent(index), this._heap[index]);

    while (condition(currentIndex)) {
      this.swap(currentIndex, this.getParentIndex(currentIndex));
      currentIndex = this.getParentIndex(currentIndex);
    }
  }

  /**
   * @param {Number} customStartIndex
   */
  heapifyDown(customStartIndex = 0) {
    /**
     * Compare parent element to its children and swap it with the appropriate child:
     * smallest child for MinHeap and largest child for MaxHeap
     * Do the same for next children after swap.
     */
    let currentIndex = customStartIndex;
    let nextIndex = null;

    while (this.hasLeftChild(currentIndex)) {
      if (this.hasRightChild(currentIndex) && this.pairIsInCorrectOrder(this.rightChild(currentIndex), this.leftChild(currentIndex))) {
        nextIndex = this.getRightChildIndex(currentIndex);
      }
      else {
        nextIndex = this.getLeftChildIndex(currentIndex);
      }

      if (this.pairIsInCorrectOrder(this._heap[currentIndex], this._heap[nextIndex])) {
        break;
      }

      this.swap(currentIndex, nextIndex);
      currentIndex = nextIndex;
    }
  }

  /**
   * Cheks if pair of heap elements is in correct order.
   * For MinHeap the first element must be always smaller or equal.
   * For MaxHeap the last element must be always bigger or equal.
   *
   * @param {*} firstElement
   * @param {*} secondElement
   * @return {Boolean}
   */
  pairIsInCorrectOrder(firstElement, secondElement) {
    throw new Error(`Not implemented`);
  }

  /**
   * Print utility method
   * @return {String}
   */
  print() {
    const n = this._heap.length;
    const h = Math.floor(Math.log2(n));
    let result = '';

    if (n < 1) {
      return result;
    }

    for (let i = 0; i <= h; i++) {
      const power = 2 ** i; // 1, 2, 4, 8... max elements per row
      const spaces = '*'.repeat(h - i);
      for (let j = 0; j < power; j++) {
        const idx = j + power - 1;
        if (idx >= n) {
          break;
        }
        result += this._heap[idx] + ' ';
      }
      result += '\n';
    }
    return result;
  }
}
