import { Comparator } from '../../comparator/comparator.mjs';
import { LinkedListNode } from './linked-list-node.mjs';

export class LinkedList {
  /**
   * @param {Function} [comparatorFunction]
   */
  constructor(comparatorFunction) {
    /** @var LinkedListNode */
    this.head = null;

    /** @var LinkedListNode */
    this.tail = null;

    /** @var Comparator */
    this.compare = new Comparator(comparatorFunction);
  }

  /**
   * @param {*[]} values
   * @param {Function} [comparatorFunction]
   * @return {LinkedList}
   */
  static fromArray(values = [], comparatorFunction) {
    const list = new LinkedList(comparatorFunction);
    values.forEach(value => list.append(value));
    return list;
  }

  /**
   * @param {*} value
   * @return {LinkedList}
   */
  prepend(value) {
    // Make new node to be the head
    const newNode = new LinkedListNode(value, this.head);
    this.head = newNode;

    // if there is no tail yet, let's make the new node a tail
    if (!this.tail) {
      this.tail = newNode;
    }

    return this;
  }

  /**
   * @param {*} value
   * @return {LinkedList}
   */
  append(value) {
    const newNode = new LinkedListNode(value);

    // If there is no head yet, let's make new node a head.
    if (!this.head) {
      this.head = newNode;
      this.tail = newNode;
      return this;
    }

    // Attach new node to the end of linked list.
    this.tail.next = newNode;
    this.tail = newNode;

    return this;
  }

  /**
   * @param {*} value
   * @return {LinkedListNode}
   */
  delete(value) {
    if (!this.head) {
      return null;
    }

    let deletedNode = null;

    // if the head must be deleted, then make 2nd node to be the head
    while (this.head && this.compare.eq(this.head.value, value)) {
      deletedNode = this.head;
      this.head = this.head.next;
    }

    let currentNode = this.head;

    if (currentNode !== null) {
      // If next node must be deleted, then make next node to be a next one
      // We will never reach the tail, becase its next is null
      while (currentNode.next) {
        if (this.compare.eq(currentNode.next.value, value)) {
          deletedNode = currentNode.next;
          currentNode.next = currentNode.next.next;
        }
        else {
          currentNode = currentNode.next;
        }
      }
    }

    // check if tail must be deleted
    if (this.compare.eq(this.tail.value, value)) {
      this.tail = currentNode;
    }

    return deletedNode;
  }

  /**
   * @param {Object} params
   * @param {*} params.value
   * @param {Function} [params.callback]
   * @return {LinkedListNode}
   */
  find({ value = undefined, callback = undefined }) {
    if (!this.head) {
      return null;
    }

    let currentNode = this.head;

    while (currentNode) {
      // If callback is specified, then try to find node by callback
      if (callback && callback(currentNode.value)) {
        return currentNode;
      }

      // If value is specified, then try to compare by value
      if (value !== undefined && this.compare.eq(currentNode.value, value)) {
        return currentNode;
      }

      currentNode = currentNode.next;
    }

    return null;
  }

  /**
   * @return {LinkedListNode}
   */
  deleteHead() {
    if (!this.head) {
      return null;
    }

    const deletedHead = this.head;

    if (this.head.next) {
      this.head = this.head.next;
    }
    else {
      this.head = null;
      this.tail = null;
    }

    return deletedHead;
  }

  /**
   * @return {LinkedListNode}
   */
  deleteTail() {
    if (this.head === this.tail) {
      // there is only one node in linked list
      const deletedTail = this.tail;
      this.head = null;
      this.tail = null;
      return deletedTail;
    }

    let currentNode = this.head;
    const deletedTail = this.tail;

    while (currentNode.next) {
      if (!currentNode.next.next) {
        currentNode.next = null;
      }
      else {
        currentNode = currentNode.next;
      }
    }

    this.tail = currentNode;

    return deletedTail;
  }

  /**
   * @return {LinkedList}
   */
  reverse() {
    let currentNode = this.head;
    let prevNode = null;
    let nextNode = null;

    while (currentNode) {
      nextNode = currentNode.next;
      currentNode.next = prevNode;
      prevNode = currentNode;
      currentNode = nextNode;
    }

    this.tail = this.head;
    this.head = prevNode;

    return this;
  }

  /**
   * @return {LinkedListNode[]}
   */
  toArray() {
    const nodes = [];

    let currentNode = this.head;

    while (currentNode) {
      nodes.push(currentNode);
      currentNode = currentNode.next;
    }

    return nodes;
  }

  /**
   * @param {Function} [callback]
   * @return {String}
   */
  toString(callback) {
    return this.toArray().map(node => node.toString(callback)).toString();
  }
}
